@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # thoth-python
-# Copyright(C) 2018, 2019 Fridolin Pokorny
+# Copyright(C) 2018, 2019, 2020 Fridolin Pokorny
 #
 # This program is free software: you can redistribute it and / or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,10 +20,14 @@
 import json
 import hashlib
 import logging
-import typing
+from typing import Dict
+from typing import Any
+from typing import Optional
+from typing import Iterable
+from typing import List
 from itertools import chain
 
-import contoml as toml
+import toml
 import attr
 
 from .exceptions import PipfileParseError
@@ -43,11 +47,11 @@ _DEFAULT_PIPFILE_SPEC = 6
 class PipfileMeta:
     """Parse meta information stored in Pipfile or Pipfile.lock."""
 
-    sources = attr.ib(type=dict)
-    requires = attr.ib(type=dict)
-    pipenv = attr.ib(type=dict)
-    hash = attr.ib(type=dict)
-    pipfile_spec = attr.ib(type=int)
+    sources = attr.ib(type=Dict[str, Source])
+    requires = attr.ib(type=Dict[str, Any], default=attr.Factory(dict))
+    pipenv = attr.ib(type=Optional[Dict[str, Any]], default=None)
+    hash = attr.ib(type=Optional[Dict[str, Any]], default=None)
+    pipfile_spec = attr.ib(type=int, default=_DEFAULT_PIPFILE_SPEC)
 
     @classmethod
     def from_dict(cls, dict_: dict):
@@ -78,7 +82,7 @@ class PipfileMeta:
         _LOGGER.debug("Generating Pipfile%s metadata section", "" if not is_lock else ".lock")
         sources_dict = [source.to_dict() for source in self.sources.values()]
 
-        result = {}
+        result = {}  # type: Dict[str, Any]
         if is_lock:
             # Pipenv is omitted.
             result["sources"] = sources_dict
@@ -104,7 +108,7 @@ class PipfileMeta:
         result = ""
 
         primary_index_added = False
-        for source in self.sources.items():
+        for source in self.sources.values():
             if not primary_index_added:
                 result += f"-i {source.url}\n"
                 primary_index_added = True
@@ -140,8 +144,8 @@ class PipfileMeta:
 class _PipfileBase:
     """A base class encapsulating logic of Pipfile and Pipfile.lock."""
 
-    packages = attr.ib(type=dict)
-    dev_packages = attr.ib(type=dict)
+    packages = attr.ib(type=Packages)
+    dev_packages = attr.ib(type=Packages)
     meta = attr.ib(type=PipfileMeta)
 
     # I wanted to reuse pipenv implementation, but the implementation is not that reusable. Also, we would like
@@ -172,7 +176,7 @@ class _PipfileBase:
         # First add index configuration.
         requirements_file = self.meta.to_requirements_index_conf()
 
-        for package_version in self.packages.items() if not develop else self.dev_packages.items():
+        for package_version in self.packages.packages.values() if not develop else self.dev_packages.packages.values():
             requirements_file += f"{package_version.name}{package_version.version}"
 
         return requirements_file
@@ -189,7 +193,7 @@ class _PipfileBase:
         _LOGGER.debug("Checking source indexes used")
 
         def _index_check(package_version: PackageVersion, source: Source):
-            if source is package_version.index:
+            if source is package_version.index or package_version.index is None:
                 return
 
             if source.name == package_version.index.name and source.url != package_version.index.url:
@@ -216,11 +220,133 @@ class _PipfileBase:
             else:
                 _index_check(package_version, self.meta.sources[package_version.index.name])
 
+    @staticmethod
+    def _construct_requirements_packages(packages: Iterable[PackageVersion]) -> str:
+        """Construct a requirements.txt/in string entry for each package."""
+        result = ""
+
+        for package in packages:
+            result += package.name
+            if package.extras:
+                result += f"[{','.join(e for e in sorted(package.extras))}]"  # type: ignore
+            if package.version and package.version != "*":
+                result += package.version
+
+            if package.markers:
+                result += "; " + package.markers
+
+            if package.hashes:
+                result += " \\\n"
+                for idx, digest in enumerate(package.hashes):
+                    result += f"    --hash={digest}"
+                    if idx != len(package.hashes) - 1:
+                        result += " \\\n"
+                    else:
+                        result += "\n"
+            else:
+                result += "\n"
+
+        return result
+
+    @classmethod
+    def _construct_requirements(
+        cls,
+        packages: Iterable[PackageVersion],
+        dev_packages: Optional[Iterable[PackageVersion]],
+        meta: PipfileMeta,
+        with_header: bool = True,
+    ) -> str:
+        result = ""
+        if with_header:
+            result = """#
+# This file is autogenerated by Thoth and is meant to be used with pip-compile
+# as provided by pip-tools.
+#
+"""
+
+        for idx, index in enumerate(meta.sources.values()):
+            if idx == 0:
+                result += f"--index-url {index.url}\n"
+            else:
+                result += f"--extra-index-url {index.url}\n"
+
+        result += "\n"
+        result += cls._construct_requirements_packages(packages)
+
+        if dev_packages:
+            result += "\n#\n# dev packages\n#\n"
+            result += cls._construct_requirements_packages(dev_packages)
+
+        return result
+
+    def construct_requirements_txt(self) -> str:
+        """Construct requirements.txt file."""
+        return self._construct_requirements(
+            self.packages.packages.values(),
+            self.dev_packages.packages.values(),
+            self.meta,
+        )
+
+
+@attr.s(slots=True)
+class ThothPipfileSection:
+    """Thoth specific section in Pipfile."""
+
+    allow_prereleases = attr.ib(type=Dict[str, bool], default=attr.Factory(dict))
+    disable_index_adjustment = attr.ib(type=bool, default=False)
+
+    def to_dict(self) -> Dict[str, Any]:
+        """Get a dict representation of Thoth specific section in Pipfile."""
+        result = attr.asdict(self)
+
+        # Keep the Thoth section minimal.
+        if not result["allow_prereleases"]:
+            result.pop("allow_prereleases")
+
+        if not result["disable_index_adjustment"]:
+            result.pop("disable_index_adjustment")
+
+        return result
+
+    @classmethod
+    def from_dict(cls, dict_: Dict[str, Any]) -> "ThothPipfileSection":
+        """Convert Thoth specific section in Pipfile to a dictionary representation."""
+        dict_ = dict(dict_)
+        allow_prereleases = dict_.pop("allow_prereleases", {})
+        disable_index_adjustment = dict_.pop("disable_index_adjustment", False)
+
+        if dict_:
+            _LOGGER.warning("Unknown entry in Thoth specific Pipfile section: %r", dict_)
+
+        if not isinstance(allow_prereleases, dict):
+            _LOGGER.warning(
+                "allow_prereleases expected to be a dictionary, but got %r instead - ignoring", type(allow_prereleases)
+            )
+        if not isinstance(disable_index_adjustment, bool):
+            _LOGGER.warning(
+                "disable_index_adjustment expected to be a boolean, but got %r instead - ignoring",
+                type(disable_index_adjustment),
+            )
+
+        for k, v in list(allow_prereleases.items()):
+            if not isinstance(k, str) or not k:
+                _LOGGER.warning("allow_prereleases expects package names, but got %r - ignoring", k)
+                allow_prereleases.pop(k)
+                continue
+            if not isinstance(v, bool):
+                _LOGGER.warning("allow_prereleases expects a boolean flag for package %r but got %r - ignoring", k, v)
+                allow_prereleases.pop(k)
+                continue
+
+        return cls(allow_prereleases=allow_prereleases, disable_index_adjustment=disable_index_adjustment)
+
 
 @attr.s(slots=True)
 class Pipfile(_PipfileBase):
     """A Pipfile representation - representation of direct dependencies of an application."""
 
+    thoth = attr.ib(type=ThothPipfileSection, default=attr.Factory(ThothPipfileSection))
+
     @property
     def data(self):
         """Return data used to compute hash based on Pipfile stored in Pipfile.lock."""
@@ -230,7 +356,7 @@ class Pipfile(_PipfileBase):
         return {"default": self.packages.to_pipfile(), "develop": self.dev_packages.to_pipfile(), "_meta": meta}
 
     @classmethod
-    def from_package_versions(cls, packages: typing.List[PackageVersion], meta: PipfileMeta = None):
+    def from_package_versions(cls, packages: List[PackageVersion], meta: PipfileMeta = None) -> "Pipfile":
         """Construct Pipfile from provided PackageVersion instances."""
         return cls(
             packages=Packages.from_package_versions([pv for pv in packages if not pv.develop], develop=False),
@@ -239,7 +365,7 @@ class Pipfile(_PipfileBase):
         )
 
     @classmethod
-    def from_file(cls, file_path: str = None):
+    def from_file(cls, file_path: str = None) -> "Pipfile":
         """Parse Pipfile file and return its Pipfile representation."""
         file_path = file_path or "Pipfile"
         _LOGGER.debug("Loading Pipfile from %r", file_path)
@@ -247,13 +373,12 @@ class Pipfile(_PipfileBase):
             return cls.from_string(pipfile_file.read())
 
     @classmethod
-    def from_string(cls, pipfile_content: str):
+    def from_string(cls, pipfile_content: str) -> "Pipfile":  # type: ignore
         """Parse Pipfile from its string representation."""
         _LOGGER.debug("Parsing Pipfile toml representation from string")
         try:
             parsed = toml.loads(pipfile_content)
-            parsed = parsed.primitive
-        except Exception as exc:
+        except Exception as exc:  # noqa: F841
             # We are transparent - Pipfile can be eigher TOML or JSON - try to parse any of these.
             try:
                 parsed = json.loads(pipfile_content)
@@ -263,11 +388,12 @@ class Pipfile(_PipfileBase):
         return cls.from_dict(parsed)
 
     @classmethod
-    def from_dict(cls, dict_):
+    def from_dict(cls, dict_) -> "Pipfile":
         """Retrieve instance of Pipfile from its dictionary representation."""
         _LOGGER.debug("Parsing Pipfile")
         packages = dict_.pop("packages", {})
         dev_packages = dict_.pop("dev-packages", {})
+        thoth_section = ThothPipfileSection.from_dict(dict_.pop("thoth", {}))
 
         # Use remaining parts - such as requires, pipenv configuration and other flags.
         meta = PipfileMeta.from_dict(dict_)
@@ -275,6 +401,7 @@ class Pipfile(_PipfileBase):
             packages=Packages.from_pipfile(packages, develop=False, meta=meta),
             dev_packages=Packages.from_pipfile(dev_packages, develop=True, meta=meta),
             meta=meta,
+            thoth=thoth_section,
         )
 
     def to_dict(self) -> dict:
@@ -282,6 +409,11 @@ class Pipfile(_PipfileBase):
         _LOGGER.debug("Generating Pipfile")
         result = {"packages": self.packages.to_pipfile(), "dev-packages": self.dev_packages.to_pipfile()}
         result.update(self.meta.to_dict())
+
+        thoth_section = self.thoth.to_dict()
+        if thoth_section:
+            result["thoth"] = thoth_section
+
         return result
 
     def to_string(self) -> str:
@@ -289,6 +421,14 @@ class Pipfile(_PipfileBase):
         _LOGGER.debug("Converting Pipfile to toml")
         return toml.dumps(self.to_dict())
 
+    def construct_requirements_in(self) -> str:
+        """Construct requirements.in file for the current project."""
+        return self._construct_requirements(
+            self.packages.packages.values(),
+            self.dev_packages.packages.values(),
+            self.meta,
+        )
+
     def to_file(self) -> None:
         """Convert the current state of Pipfile to actual Pipfile file stored in CWD."""
         with open("Pipfile", "w") as pipfile:
@@ -307,20 +447,22 @@ class Pipfile(_PipfileBase):
 class PipfileLock(_PipfileBase):
     """A Pipfile.lock representation - representation of fully pinned down stack with info such as hashes."""
 
-    pipfile = attr.ib(type=Pipfile)
+    pipfile = attr.ib(type=Optional[Pipfile])
 
     @classmethod
-    def from_package_versions(cls, pipfile: Pipfile, packages: typing.List[PackageVersion], meta: PipfileMeta = None):
+    def from_package_versions(
+        cls, pipfile: Pipfile, packages: List[PackageVersion], meta: PipfileMeta = None
+    ) -> "PipfileLock":
         """Construct Pipfile from provided PackageVersion instances."""
         return cls(
             pipfile=pipfile,
             packages=Packages.from_package_versions([pv for pv in packages if not pv.develop], develop=False),
             dev_packages=Packages.from_package_versions([pv for pv in packages if pv.develop], develop=True),
-            meta=meta,
+            meta=meta or PipfileMeta.from_dict({}),
         )
 
     @classmethod
-    def from_file(cls, file_path: str = None, pipfile: Pipfile = None):
+    def from_file(cls, file_path: str = None, pipfile: Pipfile = None) -> "PipfileLock":
         """Parse Pipfile.lock file and return its PipfileLock representation."""
         file_path = file_path or "Pipfile.lock"
         _LOGGER.debug("Loading Pipfile.lock from %r", file_path)
@@ -328,7 +470,7 @@ class PipfileLock(_PipfileBase):
             return cls.from_string(pipfile_file.read(), pipfile)
 
     @classmethod
-    def from_string(cls, pipfile_content: str, pipfile: Pipfile):
+    def from_string(cls, pipfile_content: str, pipfile: Optional[Pipfile]) -> "PipfileLock":  # type: ignore
         """Parse Pipfile.lock from its string content."""
         _LOGGER.debug("Parsing Pipfile.lock JSON representation from string")
         try:
@@ -339,7 +481,7 @@ class PipfileLock(_PipfileBase):
         return cls.from_dict(parsed, pipfile)
 
     @classmethod
-    def from_dict(cls, dict_: dict, pipfile: Pipfile):
+    def from_dict(cls, dict_: dict, pipfile: Optional[Pipfile]) -> "PipfileLock":
         """Construct PipfileLock class from a parsed JSON representation as stated in actual Pipfile.lock."""
         _LOGGER.debug("Parsing Pipfile.lock")
         meta = PipfileMeta.from_dict(dict_["_meta"])
